
# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# Dec. 2014, Boston, MA 
# April. 2014, Boston, MA


import numpy
from PIL import Image
import nibabel
import json 
import DisplayNode 
import copy

from occiput.Visualization import MultipleVolumes, MultipleVolumesNiftyPy, VolumeRenderer
from occiput.Visualization import ipy_table, has_ipy_table
from occiput.Core import transformations as tr 
from occiput import global_settings 
from occiput.global_settings import printoptions
from occiput.Core.NiftyPy_wrap import transform_grid, grid_from_box_and_affine, resample_image_on_grid, INTERPOLATION_LINEAR, INTERPOLATION_POINT
from occiput.Core.Conversion import nipy_to_occiput, nifti_to_occiput, occiput_to_nifti, occiput_from_array





class Transform_Affine(object): 
    def __init__(self,data=None,map_from="", map_to=""): 
        self.set_map_from(map_from)
        self.set_map_to(map_to)
        self.set_data(data)

    def can_inverse_left_multiply(self,obj): 
        return self.__space_of_obj(obj) == self.map_to

    def can_left_multiply(self,obj):       
        return self.__space_of_obj(obj) == self.map_from
        
    def left_multiply(self,obj): 
        # if obj is a nd-array (of size [4xN]): 
        if isinstance(obj,numpy.ndarray): 
            return numpy.dot(self.data,obj) 
        # verify if object and affine are in the same space
        if not self.can_left_multiply(obj): 
            if hasattr(obj,'map_from'): 
                X = obj.map_from
            else: 
                X = "X"
            print "Affine is incompatible with the given object: composing [%s,%s] with [%s,%s] is not allowed. "%( self.map_to,self.map_from,self.__space_of_obj(obj),X )
        data = numpy.dot(self.data,obj.data)
        # return an affine transformation is the input was an affine transformation 
        if isinstance(obj,Transform_Affine): 
            return Transform_Affine(data=data, map_from=obj.map_from, map_to=self.map_to)
        else:
            return data

    def __space_of_obj(self,obj): 
        if hasattr(obj,'space'): 
            return obj.space
        elif hasattr(obj,'map_to'): 
            return obj.map_to
        else: 
            return ""

    def export_dictionary(self): 
        return {'map_to':self.map_to, 'map_from':self.map_from, 'data':self.data.tolist()}

    def export_json(self): 
        return json.dumps(self.export_dictionary())

    def save_to_file(self,filename): 
        with open(filename, 'w') as fid: 
            json.dump(self.export_json(), fid)

    def load_from_file(self,filename): 
        with open(filename, 'r') as fid: 
            self.import_json(json.load(fid)) 

    def import_json(self,string): 
        self.import_dictionary(json.loads(string))
        
    def import_dictionary(self,dict):
        self.map_to = dict['map_to']
        self.map_from = dict['map_from']
        self.data  = numpy.asarray(dict['data'])

    def is_6DOF(self):
        #tra, rot, axis = self.to_translation_rotation() 
        #if tra is None: 
        #    return False
        #mat = numpy.dot(tr.translation_matrix(tra), tr.rotation_matrix(rot,axis)) 
        #return tr.is_same_transform(mat,self.data)
        return True # FIXME
        
    # FIXME: implement the following functions (see is_6DOF() )
    def is_9DOF(self): 
        tra, scale, rot, rot_axis = self.to_translation_scale_rotation()
        if tra is None: 
            return False   
        tra_mat = tr.translation_matrix(tra)
        scale_mat = numpy.diag([scale[0],scale[1],scale[2],1.0])
        rot_mat = tr.rotation_matrix(rot,rot_axis)
        mat = numpy.dot(tra_mat, numpy.dot(scale_mat, rot_mat ) )
        return tr.is_same_transform(mat,self.data)

    def is_rigid(self):
        #return self.is_9DOF()
        return True

    def is_rotation(self): 
        return False # FIXME

    def is_translation(self):
        return False # FIXME

    def is_scale(self): 
        return False # FIXME

    def determinant(self): 
        try: 
            det = numpy.linalg.det(self.data)
        except: 
            det = None    
        return det

    def to_translation_rotation(self): 
        tra = tr.translation_from_matrix(self.data)
        mat = self.data.copy()
        mat[0:3,3]=0
        try: 
            rot,axis,point = tr.rotation_from_matrix(mat)
        except ValueError: 
            #print "Not a rotation matrix. "
            return None, None, None
        return tra, rot, axis

    def from_translation_rotation(self, tra, rot, axis=[0,0,1]): 
        return numpy.dot(tr.translation_matrix(tra), tr.rotation_matrix(rot,axis) )

    def to_translation_scale_rotation(self): 
        # FIXME
        mat = self.data.copy()
        tra  = tr.translation_from_matrix(mat)
        tra_mat = tr.translation_matrix(tra)
        mat = numpy.dot(numpy.linalg.inv(tra_mat), mat)
        factor, origin, direction = tr.scale_from_matrix(mat)
        scale_mat = tr.scale_matrix(factor, origin, direction)
        scale = numpy.diag(scale_mat) 
        mat = numpy.dot(numpy.linalg.inv(scale_mat), mat) 
        try: 
            rot,rot_axis,point = tr.rotation_from_matrix(mat) 
        except ValueError: 
            #print "Not a rotation matrix. "
            return None, None, None, None
        #print "Rotation axis: ",rot_axis
        return tra, scale, rot, rot_axis
 
    def to_quaternion_translation(self): 
        tra = tr.translation_from_matrix(self.data)
        qua = tr.quaternion_from_matrix(self.data)
        return qua, tra 

    def from_quaternion_translation(self,quaternion,translation): 
        rot = quaternion_matrix(quaternion)
        return numpy.dot(translation, rot)

    def derivative_parameters(self, gradient_transformed_image ): 
        pass 
        # FIXME: implement (implemented in case of 6DOF in the derived class Transform_6DOF)
        
    def __get_inverse(self):
        inverse = Transform_Affine(data = self.__inverse, map_to = self.map_from, map_from = self.map_to)
        return inverse

    def __compute_inverse(self):
        self.__inverse = numpy.linalg.inv(self.data)

    def __get_data(self):
        return self.__data
    
    def set_data(self,data): 
        if not isinstance(data, numpy.ndarray): 
            if data  is None: 
                data = numpy.eye(4)
            else:
                # FIXME: raise exception
                return 
        if not data.size == 16: 
            # FIXME: raise exception 
            return 
        self.__data = data 
        self.__compute_inverse() 

    def __get_map_to(self):
        return self.__map_to
    
    def set_map_to(self, map_to): 
        self.__map_to = map_to
    
    def __get_map_from(self):
        return self.__map_from
    
    def set_map_from(self, map_from): 
        self.__map_from = map_from

    def __repr__(self): 
        with printoptions(precision=4, suppress=True):
            s = "Transformation "+"\n-from: "+str(self.map_from)+"\n-to:   "+str(self.map_to)+"\n-matrix: \n"+self.__data.__repr__()
        return s

    def _repr_html_(self):
        if not has_ipy_table: 
            return "Please install ipy_table."
        s = 'Transformation'
        table_data = [[s,'','','','','',''],
                      ['map_from','map_to','determinant','is_rigid','is_6DOF','is_rotation','is_translation'],
                      [self.map_from, self.map_to, str(self.determinant()), self.is_rigid(), self.is_6DOF(), self.is_rotation(), self.is_translation() ], ] 
        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic')
        table = ipy_table.set_cell_style(0,0, column_span=7)
        table = ipy_table.set_cell_style(0,0, align='center') 
        table = ipy_table.set_row_style(1, color='#F7F7F7')
        table = ipy_table.set_row_style(2, color='#F7F7F7')
        table = ipy_table.set_row_style(0, color='#FFFF99')
        table = ipy_table.set_row_style(1, bold = True)
        table = ipy_table.set_row_style(1, align = 'center')
        table = ipy_table.set_row_style(1, height = 30)
        table = ipy_table.set_global_style(float_format="%3.3f") 
        table2 = ipy_table.make_table(self.data)
        table3 = ipy_table.make_table(['   '])
        table3 = ipy_table.set_row_style(0, no_border='all')
        s = '<div><style>table {float: left; margin-right:10px;}</style> %s %s %s </div>'%( table._repr_html_(), table3._repr_html_() ,table2._repr_html_() )
        return s

    data     = property(__get_data, set_data)
    map_to   = property(__get_map_to, set_map_to)
    map_from = property(__get_map_from, set_map_from)
    inverse  = property(__get_inverse )





class Transform_Identity(Transform_Affine): 
    def __init__(self,map_from="",map_to=""): 
        Transform_Affine.__init__(self, numpy.eye(4),map_from,map_to)     


#class Transform_Scale(Transform_Affine): 
#    def __init__(self, factor, origin=None, direction=None, map_from="",map_to="" ): 
#        mat = tr.scale_matrix(factor, origin, direction)
#        Transform_Affine.__init__(self, mat, map_from,map_to)
class Transform_Scale(Transform_Affine):  #FIXME: figure out what transformations.py means by scale matrix 
    def __init__(self, scale_xyz, map_from="",map_to="" ): 
        mat = numpy.diag([scale_xyz[0],scale_xyz[1],scale_xyz[2],1]) 
        Transform_Affine.__init__(self, mat, map_from, map_to)


class Transform_Rotation(Transform_Affine): 
    def __init__(self, angle, direction, point=None, map_from="",map_to=""): 
        mat = tr.rotation_matrix(angle, direction, point)
        Transform_Affine.__init__(self, mat,map_from,map_to)

    def derivative_parameters(self, gradient_transformed_image ): 
        pass 

class Transform_Translation(Transform_Affine): 
    def __init__(self, direction, map_from="",map_to=""): 
        mat = tr.translation_matrix(direction)     
        Transform_Affine.__init__(self, mat,map_from,map_to)

    def derivative_parameters(self, gradient_transformed_image ): 
        pass

class Transform_6DOF(Transform_Affine): 
    def __init__(self, translation, rotation_angle, rotation_direction, rotation_point=None, map_from="",map_to=""): 
        rot = tr.rotation_matrix(rotation_angle, rotation_direction, rotation_point) 
        tra = tr.translation_matrix(translation) 
        mat = numpy.dot(tra,rot)
        Transform_Affine.__init__(self, mat,map_from,map_to) 

    def derivative_parameters(self, gradient_transformed_image, grid_transformed_image ): 
        dRx = []
        dRy = []
        dRz = []
        return [0,0,0,0,0,0]
        






class GridND(object):
    def __init__(self, data=None, space="", is_uniform=True, is_affine=True, is_axis_aligned=True): 
        self.ndim = None
        self.__set_data(data)
        self.space = space 
        self.__clear_cache() 
        self.__is_uniform = is_uniform 
        self.__is_affine  = is_affine 
        self.__is_axis_aligned = is_axis_aligned 

    def min(self): 
        if self.__min  is None: 
            self.__min = eval('self.data%s'%('.min(0)'*self.ndim))
        return self.__min

    def max(self): 
        if self.__max  is None:
            self.__max = eval('self.data%s'%('.max(0)'*self.ndim))
        return self.__max 

    def span(self): 
        if self.__span  is None:
            self.__span = self.max() - self.min() 
        return self.__span 

    def center(self,use_corners_only=True): 
        if self.__center  is None or (self.__use_corners_only != use_corners_only): 
            if use_corners_only: 
                corners = self.corners()
                center = corners.mean(1)
            else: 
                center = None 
                #FIXME: implement 
            self.__center = center
            self.__use_corners_only = use_corners_only
        return self.__center

    def mean_distance_from_point(self, point, use_corners_only=True): 
        if use_corners_only: 
            corners = self.corners()
            dist = corners - numpy.tile( numpy.asarray(point).reshape(3,1),[1,corners.shape[1]] ) 
            dist =  numpy.sqrt((dist * dist)).sum(1) / corners.shape[1]
        else:
            dist = None
            #FIXME: implement 
        return dist 

    def mean_dist_from_center(self): 
        if self.__mean_dist_center  is None: 
            center = self.center() 
            self.__mean_dist_center = self.mean_distance_from_point(center) 
        return self.__mean_dist_center
        
    def get_shape(self):
        return self.data.shape

    def is_uniform(self): 
        """Returns True if the grid is uniform. """
        return self.__is_uniform
        #FIXME: change the flags when grid is transformed 
    
    def is_affine(self): 
        """Returns True if the grid is the affine transformation of a uniform grid. """
        return self.__is_affine 

    def is_axis_aligned(self): 
        """Returns True if the grid is uniform and aligned to the x,y,z axis. """
        return self.__is_axis_aligned 

    def transform(self, affine_from_grid): 
        return transform_grid(self, affine_from_grid )

    def corners(self, homogeneous_coords=False): 
        if self.__corners  is None: 
            n_corners = 2**self.ndim 
            corners = []
            for i in range(n_corners): 
                b = eval( '['+bin(i)[2:].zfill(self.ndim).replace('0','0,').replace('1','1,') +"]" )
                b = (numpy.asarray(self.data.shape[0:self.ndim])-1) * numpy.asarray(b)
                s = str(b.tolist())[1:-1]
                corner = eval('self.data[%s,:]'%s)
                corners.append(corner)
            corners = numpy.asarray(corners).transpose() 
            if homogeneous_coords: 
                corners2 = numpy.ones((4,n_corners))
                corners2[0:3,:] = corners
                corners = corners2
            self.__corners = corners
        return self.__corners

    def __get_data(self):
        return self.__data 
    
    def __set_data(self,data): 
        self.__data = data
        if data is not None: 
            self.ndim = self.data.ndim-1 
        else: 
            self.ndim = None
        self.__clear_cache() 
        
    def __get_shape(self): 
        return self.data.shape 
    
    def __clear_cache(self): 
        self.__min  = None 
        self.__max  = None 
        self.__span = None
        self.__center = None
        self.__use_corners_only = None
        self.__mean_dist_center = None
        self.__corners = None

    def __repr__(self):  
        with printoptions(precision=4):
            if self.is_axis_aligned():
                type = 'axis aligned'
            elif self.is_affine():
                type = 'affine'
            elif self.is_uniform():
                type = 'uniform'
            else: 
                type = ''
            if type!='': 
                type = " (%s)"%type
            s = "Grid%dD %s: "%(int(self.ndim),type)
            s = s + "\n - space -------------------> " + str(self.space)
            s = s + "\n - shape -------------------> " + str(list(self.shape)) 
            s = s + "\n - min ---------------------> " + str(self.min()) 
            s = s + "\n - max ---------------------> " + str(self.max()) 
            s = s + "\n - span --------------------> " + str(self.span()) 
            s = s + "\n - center ------------------> " + str(self.center()) 
            s = s + "\n - mean dist from center: --> " + str(self.mean_dist_from_center()) 
            s = s + "\n"
        return s
    
    def _repr_html_(self): 
        if not has_ipy_table: 
            return "Please install ipy_table."
        if self.is_axis_aligned():
            type = 'axis aligned'
        elif self.is_affine():
            type = 'affine'
        elif self.is_uniform():
            type = 'uniform'
        else: 
            type = ''
        if type!='': 
            type = " (%s)"%type
        s = "Grid%dD %s: "%(int(self.ndim),type)
        def pretty(list):
            return str(['{0:.2f}'.format(flt) for flt in list])
        table_data = [[s,'','','','','',''],
                      ['space','shape','min','max','span','center','spread'],
                      [str(self.space), pretty(self.shape) ,pretty(self.min()) ,pretty(self.max()) ,pretty(self.span()) ,pretty(self.center()) ,pretty(self.mean_dist_from_center())], ] 
        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic')
        table = ipy_table.set_cell_style(0,0, column_span=7)
        table = ipy_table.set_cell_style(0,0, align='center') 
        table = ipy_table.set_row_style(1, color='#F7F7F7')
        table = ipy_table.set_row_style(2, color='#F7F7F7')
        table = ipy_table.set_row_style(0,color='#CCFF99')
        table = ipy_table.set_row_style(1, bold = True)
        table = ipy_table.set_row_style(1, align = 'center')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        s = table._repr_html_() 
        return s

    data = property(__get_data, __set_data )
    shape = property(__get_shape )

    
class Grid3D(GridND):
    def __init__(self, data=None, space=""): 
        GridND.__init__(self, data, space)
        self.ndim = 3






class ImageND(object): 
    def __init__(self, data=None, affine=None, space="", mask_flag=0):
        self.ndim   = None
        self.space = ""
        if isinstance(data,str): 
            self.load_from_file(data)
        else: 
            self.set_data(data)
        self.set_affine(affine)
        self.set_space(space)

        self.background = global_settings.get_default_background() 
        self.use_gpu    = global_settings.is_gpu_enabled() 
        
        self.set_mask_flag(mask_flag)

    def set_mask_flag(self, is_mask): 
        self.__is_mask = is_mask
    
    def is_mask(self): 
        """Returns True if the image is a mask (i.e. the values should be interpreted as discrete labels - take note 
        that, however, that this reports the state of a flag; the data can be non-integer. )"""
        return self.__is_mask 

    def set_lookup_table(self,lut): 
        """Set a lookup table for visualization. """
        self.__lut = lut 

    def get_lookup_table(self): 
        try: 
            lut = self.__lut
        except: 
            lut = None 
        return lut

    def set_data(self,data): 
        if isinstance(data,numpy.ndarray): 
            self.data = data
        else:   
            self.data = numpy.asarray([]) 
        self.ndim = self.data.ndim

    def min(self): 
        return self.data.min() #FIXME: pass on optional parameters
    
    def max(self): 
        return self.data.max() #FIXME: pass on optional parameters

    def set_affine(self,affine): 
        if isinstance(affine,Transform_Affine): 
            self.affine = affine
        elif isinstance(affine,numpy.ndarray): 
            self.affine = Transform_Affine(data=affine)
        elif affine is None: 
            self.affine = Transform_Identity()  
        else: 
            print "'affine' must be an instance of Affine"
        self.affine.map_to = self.space        
        
    def set_space(self,space): 
        self.affine.map_to = space
        self.space = space 
        self.affine.map_from = "index" 

    def get_shape(self): 
        return self.data.shape 

    def get_world_grid(self, n_points=None): 
        if n_points  is None: 
            n_points = self.get_shape() 
        grid = grid_from_box_and_affine(self.get_world_grid_min(), self.get_world_grid_max(), n_points) 
        return grid

    def get_world_grid_min(self): 
        d = self.get_data()
        s = numpy.asarray(d.shape)-1
        corners = numpy.asarray([ [0,0,0,1], [s[0],0,0,1], [0,s[1],0,1], [s[0],s[1],0,1], [0,0,s[2],1], [s[0],0,s[2],1], [0,s[1],s[2],1], [s[0],s[1],s[2],1] ]).transpose()
        corners = self.affine.left_multiply(corners)
        corners=corners[0:3,:]
        m = corners.min(1)
        return m
         
    def get_world_grid_max(self): 
        d = self.get_data()
        s = numpy.asarray(d.shape)-1
        corners = numpy.asarray([ [0,0,0,1], [s[0],0,0,1], [0,s[1],0,1], [s[0],s[1],0,1], [0,0,s[2],1], [s[0],0,s[2],1], [0,s[1],s[2],1], [s[0],s[1],s[2],1] ]).transpose()
        corners = self.affine.left_multiply(corners)
        corners=corners[0:3,:]
        M = corners.max(1)
        return M

    def get_pixel_grid(self): 
        print "get_pixel_grid"
        n_points = numpy.uint32(self.shape)
        grid = grid_from_box_and_affine(numpy.asarray([0,0,0]), n_points-1, n_points) 
        return grid

    def transform(self,affine): 
        if not isinstance(affine,Transform_Affine): 
            print "Transformation must be an instance of Affine."
            # FIXME raise error
            return None
        self.affine = affine.left_multiply(self.affine)
        #return self 

    def save_to_file(self, filename): 
        nii = occiput_to_nifti(self)
        nibabel.save(nii,filename)

    def __get_shape(self):
        return self.data.shape 
        
    def __get_size(self): 
        return self.data.size 

    def copy(self): 
        return copy.copy(self)

    shape = property(__get_shape)
    size  = property(__get_size)


class Image3D(ImageND):
    def __init__(self, data=None, affine=None, space=""): 
        ImageND.__init__(self, data, affine, space)
        self.ndim = 3 

    def compute_resample_on_grid(self,grid,affine_grid_to_world=None, verify_mapping=True,interpolation_mode=INTERPOLATION_LINEAR): 
        resampled_data = resample_image_on_grid(self, grid, affine_grid_to_world, verify_mapping, self.background, self.use_gpu, interpolation_mode)
        # create new Image3D object
        return Image3D(data=resampled_data) #FIXME: perhaps just return the raw resampled data
    
    def compute_resample_in_box(self,box_min,box_max,box_n): 
        pass 

    def compute_resample_in_space_of_image(self,image): 
        pass 

    def compute_gradient_on_grid(self, grid, affine_grid_to_world=None, verify_mapping=True): 
        resampled_data = resample_image_on_grid(self, grid, affine_grid_to_world, verify_mapping, self.background, self.use_gpu)
        # create new Image3D object
        #print resampled_data.max()
        gradient_data = numpy.gradient(resampled_data) #FIXME: use NiftyPy
        return gradient_data
        
    def compute_gradient_in_box(self,box): 
        pass 
    
    def compute_gradient_in_space_of_image(self,image): 
        pass 

    def compute_gradient_pixel_space(self):
        gradient_data = numpy.gradient(self.data)  #FIXME: use NiftyPy
        return gradient_data

    def compute_smoothed(self, smoothing): 
        pass 

    def get_data(self):
        return self.data 

    def export_image(self,index,axis=0,normalise=True,scale_factor=None,shrink=None,rotate=0): 
        pass 

    def has_data(self): 
        return self.data is not None

    def min(self): 
        return self.data.min()
    
    def max(self): 
        return self.data.max()

    def display_in_browser(self, axis=0): 
        D = self.display(axis) 
        D.display_in_browser() 

    def display(self, axis=0,shrink=None,rotate=None,subsample_slices=None,scales=None,open_browser=None): 
        # The following is a quick fix: use MultipleVolumesNiftyPy if the image has small size, 
        # MultipleVolumes otherwise. MultipleVolumesNiftyPy makes use of the GPU but crashes with large images. 
        # NOTE that MultipleVolumes produces a different visualisation from MultipleVolumesNiftyPy: 
        # it displays the raw imaging data, without accounting for the transformation to world space. 
        # Modify MultipleVolumesNiftyPy so that it processes the data in sequence if it is too large for the GPU. 
        # Then get rid of  MultipleVolumes. 
        #if self.size <= 256**3: 
        #    D = MultipleVolumesNiftyPy([self],axis,open_browser=open_browser) 
        #else: 
        D = MultipleVolumes([self],axis=axis,shrink=shrink,rotate=rotate,subsample_slices=subsample_slices,scales=scales,open_browser=open_browser)
        return D 

    def display_with(self, other_images, axis=0,shrink=None,rotate=None,subsample_slices=None,scales=None,open_browser=None): 
        if isinstance(other_images, type( () )): 
            other_images = list(other_images)
        elif isinstance(other_images, type( [] )): 
            other_images = other_images
        else: 
            other_images = [other_images,]
        D = MultipleVolumes([self,]+other_images,axis=axis,shrink=shrink,rotate=rotate,subsample_slices=subsample_slices,scales=scales,open_browser=open_browser)
        return D 

    def display_slice(self,axis=0,index=None,open_browser=False): 
        if index is None: 
            if axis==0: 
                index = numpy.uint32(self.shape[0]/2)
            if axis==1: 
                index = numpy.uint32(self.shape[1]/2)
            if axis==2: 
                index = numpy.uint32(self.shape[2]/2)
            if axis==0: 
                a = self.data[index,:,:].reshape((self.shape[1],self.shape[2]))  
            if axis==1: 
                a = self.data[:,index,:].reshape((self.shape[0],self.shape[2]))  
            if axis==2: 
                a = self.data[:,:,index].reshape((self.shape[0],self.shape[1]))          
        D = DisplayNode.DisplayNode()
        im = Image.fromarray(a).convert("RGB").rotate(90)
        D.display('image', im, open_browser) 
        return D 
    
    def volume_render(self,scale=1.0): 
        V = VolumeRenderer(self,scale=scale) 
        return V.display() 

    def _repr_html_(self): 
        display = self.display()
        if display: 
            return display._repr_html_()


    # Overload math operators
    def _is_same_type(self, other): 
        return isinstance(other,self.__class__)

    def _is_in_same_space(self, other): 
        # FIXME: implement
        return True 

    def _is_on_same_grid(self, other): 
        # FIXME: implement
        return True 

    def __add__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = out.data+other.data 
                    return out
                else: 
                    #FIXME: implement 
                    print "SUM of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "SUM of images not in the same space. It cannot be done. "
        else: 
            out=self.copy()
            out.data = out.data+other
            return out
        return None 

    def __sub__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = out.data-other.data 
                    return out
                else: 
                    #FIXME: implement 
                    print "SUB of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "SUB of images not in the same space. It cannot be done. "
        else: 
            out=self.copy()
            out.data = out.data-other
            return out
        return None
    
    def __mul__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = out.data*other.data 
                    return out
                else: 
                    #FIXME: implement 
                    print "MUL of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "MUL of images not in the same space. It cannot be done. "
        else: 
            out=self.copy()
            out.data = out.data*other
            return out
        return None
    
    def __div__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = out.data/other.data 
                    return out
                else: 
                    #FIXME: implement 
                    print "DIV of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "DIV of images not in the same space. It cannot be done. "
        else: 
            out=self.copy()
            out.data = out.data/other
            return out
        return None
            
    def __radd_(self, other): 
        return self.__add__(other) 

    def __rmul_(self, other): 
        return self.__mul__(other)

    def __rsub__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = other.data - out.data
                    return out
                else: 
                    #FIXME: implement 
                    print "SUB of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "SUB of images not in the same space. It cannot be done. "
        else: 
            out=self.copy()
            out.data = other-out.data
            return out
        return None

    def __rdiv__(self, other): 
        if self._is_same_type(other): 
            if self._is_in_same_space(other): 
                if self._is_on_same_grid(other): 
                    out = self.copy() 
                    out.data = other.data / out.data
                    return out
                else: 
                    #FIXME: implement 
                    print "DIV of images on different grids; the right hand side image must be resampled, please implement this."
            else: 
                # FIXME: raise error 
                print "DIV of images not in the same space. It cannot be done. "
        else: 
            out = self.copy()
            out.data = other / out.data
            return out
        return None
    








# ## Basic SPECT reconstruction example

import time
import sys
from occiput.Reconstruction.SPECT import SPECT_Static_Scan 
from occiput.DataSources.Synthetic.Shapes import uniform_spheres_ring
from pylab import *
ion()

def demo_SPECT_reconstruction(datatype="phantom"): 

    ### 1) Create an instance of a SPECT static scan
    spect = SPECT_Static_Scan()
    spect.set_use_gpu(True)
    
    ### 2) Set the geometry of the scanner 
    spect.set_n_pixels(128,128)
    spect.set_gantry_angular_positions(0.0, 360.0, 59) 
    
    ### 3) Load the projection data and the attenuation map  
    if datatype == "phantom": 
        phantom_activity    = uniform_spheres_ring([128,128,128],inner_value=1000.0) 
        phantom_attenuation = uniform_spheres_ring([128,128,128],inner_value=0.2) 
        measurement = spect.project(phantom_activity.data, attenuation=phantom_attenuation.data)
        spect.set_measurement(measurement.data) 
        spect.set_attenuation(phantom_attenuation.data)
    else: 
        spect.load_measurement_from_file('../data/spect/projection.nii')
        spect.load_attenuation_from_file('../data/spect/attenuation.nii')

    ### Let's give a look at the input data 
    figure(figsize=[18,6])
    subplot(1,2,1); imshow(spect._measurement[:,:,10],cmap='gray');
    subplot(1,2,2); imshow(spect._attenuation[64,:,:],cmap='gray');
    raw_input("Press Enter to start SPECT reconstruction...") 

    ### 4) Define the geometry of the reconstruction volume 
    spect.set_pixel_size(4.8,4.8)
    spect.set_radius(200.0)
    spect.set_psf(fwhm0_mm=5.0, depth_dependence=0.0001)



    ### 5) Reconstruction !
    
    activity = spect.estimate_activity(iterations=10, subset_size=16, subset_mode='random', method='EM') 

    figure(figsize=[18,6])
    subplot(131); imshow(activity.data[:,:,64],cmap='gray',vmax=180.0); 
    subplot(132); imshow(activity.data[:,80,:],cmap='gray',vmax=180.0); 
    subplot(133); imshow(activity.data[64,:,:],cmap='gray',vmax=180.0)

    ### 6) Save the tomogram 
    #activity.save_to_file('../data/spect/spect_01_reconstruction.nii')

    raw_input("Done! Press Enter to quit...") 

if __name__ == "__main__": 
    demo_SPECT_reconstruction("phantom")




# occiput 
# Stefano Pedemonte 
# April 2014 
# April 2015 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 


from occiput.Reconstruction.SPECT import Scintillators 
from occiput.Reconstruction.SPECT import Collimators 
from occiput.DataSources.FileSources import import_nifti 

# Import NiftyPy ray-tracers
from occiput.Core.NiftyPy_wrap import SPECT_project_parallelholes, SPECT_backproject_parallelholes, has_NiftyPy
from occiput.Core import Image3D
from occiput.Core.Errors import FileNotFound, UnknownParameter, UnexpectedParameter

from occiput.Visualization import ProgressBar, svgwrite, has_svgwrite, ipy_table, has_ipy_table
from occiput.Visualization.Colors import *

# Import DisplayNode for IPython integration
from DisplayNode import DisplayNode 

# Import various other libraries 
from numpy import *
from numpy.random import randint 
from numpy import asfortranarray, asfortranarray
from scipy import optimize 
import scipy
import scipy.signal


# Default parameters 
DEFAULT_ITERATIONS  = 20
DEFAULT_SUBSET_SIZE = 32
EPS                 = 1e-9






class SPECT_Projection(): 
    """SPECT projection object. """
    def __init__(self, data): 
        self.data = data        
    
    def get_data(self):
        """Returns the raw projection data (note that is can be accessed also as self.data ). """
        return data

    def save_to_file(self, filename): 
        h5f = h5py.File(filename, 'w')
        h5f.create_dataset('data', data=self.data)
        #h5f.create_dataset('size_x', data=size_x)
        #h5f.create_dataset('size_y', data=size_y)
        h5f.close()

    def get_integral(self): 
        return self.data.sum() 

    def to_image(self,data,index=0,scale=None,absolute_scale=False): 
        from PIL import Image
        a = float32(data[:,:,index].reshape((data.shape[0],data.shape[1])))   
        if scale is None:
            a = 255.0*(a)/(a.max()+1e-12)
        else: 
            if absolute_scale: 
                a = scale*(a) 
            else: 
                a = scale*255.0*(a)/(a.max()+1e-12)
        return Image.fromarray(a).convert("RGB")
        
    def display_in_browser(self,axial=True,azimuthal=False,index=0,scale=None): 
        self.display(axial=axial,azimuthal=azimuthal,index=index,scale=scale,open_browser=True)

    def display(self,scale=None,open_browser=False): 
        data = self.data
        d = DisplayNode()
        images = []
        progress_bar = ProgressBar(height='6px', width='100%%', background_color=LIGHT_GRAY, foreground_color=GRAY) 
        if scale is not None:
            scale = scale*255.0/(data.max()+1e-12)
        else: 
            scale = 255.0/(data.max()+1e-12) 
        N_projections = self.data.shape[2]
        N_x = self.data.shape[0]
        N_y = self.data.shape[1]
        print "SPECT Projection   [N_projections: %d   N_x: %d   N_y: %d]"%(N_projections,N_x,N_y)
        for i in range( N_projections ): 
                images.append( self.to_image(data,i,scale=scale,absolute_scale=True) ) 
                progress_bar.set_percentage(i*100.0/N_projections)                         
        progress_bar.set_percentage(100.0)
        return d.display('tipix', images, open_browser)
        
    def _repr_html_(self): 
        return self.display()._repr_html_()
        
        




class SubsetGenerator():  
    def __init__(self,N_positions):
        self._N_positions = N_positions

    def new_subset(self,mode,subset_size): 
        if mode=='random': 
            return self._random_no_replacement(subset_size) 
        elif mode=='ordered': 
            raise UnexpectedParameter("'%s' subset selection mode not yet supported."%str(mode))
        else: 
            raise UnexpectedParameter("'mode' parameter %s not recognised."%str(mode))
        
    def all_active(self): 
        return ones((self._N_positions),dtype=uint32) 

    def _random_no_replacement(self,subset_size): 
        if subset_size>=self._N_positions: 
            return self.all_active() 
        M = zeros((self._N_positions),dtype=int32) 
        n = 0
        while n<subset_size:
            active = randint(self._N_positions)
            if M[active] == 0: 
                M[active] = 1
                n+=1
        return M




def deg_to_rad(deg):
    return deg*pi/180.0

def rad_to_deg(rad):
    return rad*180.0/pi
    






class SPECT_Static_Scan(object):
    def __init__(self): 
        self._name         = "Generic SPECT Scanner"     
        self._scanner_type = "SPECT" 
        self._manufacturer = "No manufacturer"
        self._version = "0.0"
        # scanner parameters are named with 'self._p_xxx' 
        self._p_gantry_angular_positions      = 180   #[adim,integer]
        self._p_gantry_angular_position_first = 0.0   #[degrees]
        self._p_gantry_angular_position_last  = 358.0 #[degrees]     
        self._subset_generator = SubsetGenerator(self._p_gantry_angular_positions)                     
        self._p_scan_time_sec = 600.0                 #[seconds]
        self._p_radius_mm     = 300.0                 #[mm]
        self._p_n_pix_x       = 128                   #[adim]
        self._p_n_pix_y       = 128                   #[adim]
        self._p_pix_size_x_mm = 2.5                   #[mm]
        self._p_pix_size_y_mm = 2.5                   #[mm]
        self.set_background_activity(0.0) 
        self.set_background_attenuation(0.0) 
        self.set_use_gpu(True) 
        self.set_truncate_negative(False)
        self.set_scintillator( Scintillators.Ideal() )
        self.set_collimator( Collimators.LEHR() ) 
        self._measurement = None 
        self._need_update_norm = True
        self._attenuation = None 

    def get_name(self):
        return self._name

    def get_type(self):
        return self._scanner_type

    def get_manufacturer(self):
        return self._manufacturer

    def get_version(self): 
        return self._version

    def _get_parameters(self):
        parameters = {}
        dic = self.__dict__
        for k in dic.keys(): 
            if k.startswith('_p_'):
                parameters[k[3:]]=dic[k]        
        return parameters 

    def get_gantry_angular_positions(self): 
        return (self._p_gantry_angular_position_first, self._p_gantry_angular_position_last, self._p_gantry_angular_positions)

    def set_gantry_angular_positions(self, first_position_deg, last_position_deg, N_positions): 
        if not ( isscalar(first_position_deg) and isscalar(last_position_deg) and isscalar(N_positions) ): 
            raise UnexpectedParameter('Expected scalar values.')
        if not isinstance(N_positions,type(1)): 
            raise UnexpectedParameter('Expected an integer value.') 
        self._p_gantry_angular_position_first = first_position_deg
        self._p_gantry_angular_position_last =  last_position_deg
        self._p_gantry_angular_positions = N_positions
        self._subset_generator = SubsetGenerator(self._p_gantry_angular_positions)

    def get_scan_time(self): 
        return self._p_scan_time_sec

    def set_scan_time(self,scan_time_sec): 
        if not isscalar(scan_time_sec): 
            raise UnexpectedParameter('Expected a scalar value.')
        self._p_scan_time_sec = scan_time_sec

    def get_radius(self): 
        return self._p_radius_mm 

    def set_radius(self,radius_mm): 
        if not isscalar(radius_mm): 
            raise UnexpectedParameter('Expected a scalar value.') 
        self._p_radius_mm = radius_mm

    def get_n_pixels(self): 
        return (self._p_n_pix_x, self._p_n_pix_y)

    def set_n_pixels(self,n_pixels_x,n_pixels_y): 
        if (not isscalar(n_pixels_x)) or (not isscalar(n_pixels_y)): 
            raise UnexpectedParameter('Expected integer scalar values.') #FIXME: make sure it is integer 
        self._p_n_pix_x = n_pixels_x
        self._p_n_pix_y = n_pixels_y
        self._need_update_norm = True

    def get_pixel_size(self): 
        return (self._p_pix_size_x_mm, self._p_pix_size_y_mm)

    def set_pixel_size(self,pixel_size_x,pixel_size_y): 
        if (not isscalar(pixel_size_x)) or (not isscalar(pixel_size_y)): 
            raise UnexpectedParameter('Expected scalar values.') 
        self._p_pix_size_x_mm = pixel_size_x
        self._p_pix_size_y_mm = pixel_size_y

    def get_scintillator(self):
        return self._scintillator 

    def set_scintillator(self,scintillator): 
        if not isinstance(scintillator,Scintillators.BaseScintillatorSPECT): 
            raise UnexpectedParameter('Expected an instance of BaseScintillatorSPECT')
        self._scintillator = scintillator 
        self.__make_psf() 
        self._need_update_norm = True

    def get_collimator(self): 
        return self._collimator 

    def set_collimator(self,collimator): 
        if not isinstance(collimator,Collimators.BaseCollimatorSPECT): 
            raise UnexpectedParameter('Expected an instance of BaseCollimatorSPECT')
        self._collimator = collimator 
        self.__make_psf() 
        self._need_update_norm = True
        
    def set_background_activity(self,value): 
        self._background_activity    = value 
        
    def get_background_activity(self,value): 
        return self._background_activity
    
    def set_background_attenuation(self,value): 
        self._background_attenuation = value 
        
    def get_background_attenuation(self,value): 
        return self._background_attenuation
        
    def set_use_gpu(self, value): 
        self._use_gpu = value 
    
    def set_truncate_negative(self,value): 
        self._truncate_negative = value 

    def get_camera_positions(self): 
        return float32(linspace(deg_to_rad(self._p_gantry_angular_position_first),deg_to_rad(self._p_gantry_angular_position_last),self._p_gantry_angular_positions).reshape((self._p_gantry_angular_positions,1))) 

    def project(self, activity, attenuation=None, cameras=None, psf=None, subsets_array=None): 
        if isinstance(activity,ndarray): 
            activity = float32(activity)
        else: 
            activity = float32(activity.data)
        if attenuation is None:
            attenuation = self._attenuation 
            if attenuation is not None: 
                if isinstance(attenuation,ndarray):
                    attenuation = float32(attenuation)
                else: 
                    attenuation = float32(attenuation.data)
        if cameras  is None: 
            cameras = self.get_camera_positions()
        # subsets: 
        if subsets_array is not None: 
            cameras=cameras[where(subsets_array)]
        if psf  is None: 
            psf=self._psf
        proj = SPECT_project_parallelholes(activity, cameras, attenuation, psf, self._background_activity, self._background_attenuation, self._use_gpu, self._truncate_negative)
        return SPECT_Projection(proj) 


    def backproject(self, projection, attenuation=None,  cameras=None, psf=None, subsets_array=None):
        if isinstance(projection,ndarray): 
            projection = float32(projection)
        else: 
            projection = float32(projection.data)
        if attenuation is None:
            attenuation = self._attenuation 
            if attenuation is not None: 
                if isinstance(attenuation,ndarray):
                    attenuation = float32(attenuation)
                else: 
                    attenuation = float32(attenuation.data)
        if cameras  is None: 
            cameras = self.get_camera_positions()
        # subsets: 
        if subsets_array is not None: 
            cameras=cameras[where(subsets_array)]
        if psf  is None: 
            psf=self._psf   
        backproj = SPECT_backproject_parallelholes(projection, cameras, attenuation, psf, self._background_activity, self._background_attenuation, self._use_gpu, self._truncate_negative)
        return Image3D(backproj)

    def scan(self,activity_Bq,scan_time_sec=None): 
        if scan_time_sec  is None: 
            scan_time_sec = self.get_scan_time() 
        sinogram = 0
        return sinogram
        
    def __make_probabilistic_graphical_model(self): 
        pass 
        from occiput.Visualization import Graph 
        self.graph = Graph({'nodes':[{'name': 'activity', 'type': 0},{'name': 'counts', 'type': 0}],'links':[{'source': 'activity', 'target': 'counts', 'type': 't1'},]})

    def __make_psf(self): 
        self._psf = None  #FIXME 

    def set_psf(self, fwhm0_mm=0.5, depth_dependence=0.0, n_pixels=5): 
        radius = self.get_radius()
        N = self._p_n_pix_x
        [pixx,pixy] = self.get_pixel_size() 
        psf = zeros([n_pixels,n_pixels,N]) 

        def gaussian(fwhm,size): 
            x = arange(0, size, 1, float)
            y = x[:,newaxis]
            x0 = y0 = (size-1) / 2.0
            return exp(-4*log(2) * ((x-x0)**2 + (y-y0)**2) / fwhm**2)

        for i in range(N): 
            distance = radius-N/2*pixx+i*pixx
            if distance<0: 
                distance=0
            fwhm_mm = fwhm0_mm + depth_dependence*distance
            fwhm = fwhm_mm/pixy
            psf[:,:,i] = gaussian(fwhm,n_pixels) 
        self._psf = float32(psf) 

    def get_normalization(self): 
        if self._need_update_norm: 
            self._compute_normalisation() 
        return self._norm 
        
    def _compute_normalisation(self): 
        subsets_array = self._subset_generator.all_active()
        self._norm = self.backproject(ones(( self._p_n_pix_x,self._p_n_pix_y,self._p_gantry_angular_positions ),dtype=float32, order="F") ).data 
        self._need_update_norm = False 

    def set_attenuation(self, attenuation): 
        self._attenuation = attenuation 

    def load_attenuation_from_file(self,attenuation_file):
        attenuation = import_nifti(attenuation_file).data
        self.set_attenuation(attenuation)
        
    def estimate_activity(self, attenuation=None, psf=None, iterations=DEFAULT_ITERATIONS, subset_size=DEFAULT_SUBSET_SIZE, subset_mode='random', method="EM", m=32, factr=0.01, pgtol = 1e-16, maxfun = 10000, smoothing=0.0, activity=None): 
        progress_bar = ProgressBar() 
        progress_bar.set_percentage(0.1)
        if activity==None: 
            activity = ones((self._p_n_pix_x,self._p_n_pix_y,self._p_n_pix_x),dtype=float32, order="F")
        if attenuation is None:
            attenuation = self._attenuation 
            if attenuation is not None: 
                if isinstance(attenuation,ndarray):
                    attenuation = float32(attenuation)
                else: 
                    attenuation = float32(attenuation.data)
        if psf  is None: 
            psf=self._psf
        if method=="EM": 
            print "Reconstruction method: EM"
            for i in range(iterations): 
                # Subsets: 
                if subset_size is None:
                    subsets_array = None
                    subset_size = self._p_gantry_angular_positions
                elif subset_size >= self._p_gantry_angular_positions: 
                    subsets_array = None
                    subset_size = self._p_gantry_angular_positions
                else: 
                    subsets_array = self._subset_generator.new_subset(subset_mode, subset_size)
                if subsets_array is not None: 
                    proj = self.project(activity,attenuation=attenuation,psf=psf,subsets_array=subsets_array).data
                    measurement = self._measurement[:,:,where(subsets_array)].reshape((self._p_n_pix_x,self._p_n_pix_y,subset_size))
                    measurement = asfortranarray(ascontiguousarray(measurement))
                    P = (measurement+EPS)/(proj+EPS)
                    norm = self.backproject(ones(( self._p_n_pix_x,self._p_n_pix_y, subset_size),dtype=float32, order="F"), attenuation=attenuation, psf=psf, subsets_array=subsets_array).data 
                    update = (self.backproject( P, attenuation=attenuation, psf=psf, subsets_array=subsets_array).data+EPS) / (norm +EPS) 
                else: 
                    proj = self.project(activity, attenuation=attenuation, psf=psf).data
                    P = (self._measurement+EPS)/(proj+EPS)   
                    norm = self.get_normalization()  
                    update = (self.backproject( P, attenuation=attenuation, psf=psf).data+EPS) / (norm +EPS) 
                activity = activity * update #* self.get_mask().data

                progress_bar.set_percentage((i+1)*100.0/iterations) 
                #print "Iteration: %d    max act: %f    min act: %f    max proj: %f    min proj: %f    max norm: %f    min norm: %f"%(i, activity.max(), activity.min(), proj.max(), proj.min(), norm.data.max(), norm.data.min() )
            progress_bar.set_percentage(100.0)
        elif method=="LBFGS": 
            print "Reconstruction method: LBFGS-B"
            bounds = [(None,None)] * activity.size
            for i in range(0,activity.size):
                bounds[i] = (0, None)
            args = [activity.shape,smoothing]
            activity0 = float64(activity.reshape(activity.size))
           # print "SIZE ACTIVITY0: ",activity0.shape
            activity_rec,f,d = optimize.fmin_l_bfgs_b(self.get_likelihood, activity0, fprime=self.get_gradient_activity, m=m, factr=factr, pgtol=pgtol, args=args,maxfun=maxfun,iprint=0,bounds=bounds)     
            activity = float32(activity_rec.reshape(activity.shape))
            progress_bar.set_percentage(100.0)
        else: 
            raise UnexpectedParameter("Reconstruction method %s unknown"%method)
        return Image3D(activity)
            
    def get_likelihood(self,activity, activity_size, smoothing=0.0):
                """Returns the likelihood value - given the activity. This at the moment implements the Poisson likelihood only. """
                eps = 1e-16
                sinogram = self._measurement
                psf = self._psf
                attenuation = None
                sinosize = sinogram.size
                activity = activity.reshape(activity_size)
                if any(activity<0): 
                    return 0
                #if any(activity<0): 
                #    activity[activity<0]=0
                print "MIN MAX activity: ",activity.min(), activity.max()
                proj = self.project(activity).data      #FIXME: optionally use subsets 
                print "MIN MAX proj: ",proj.min(), proj.max()
                log_proj = log(proj+eps)
                print "MIN MAX log_proj: ",log_proj.min(), log_proj.max()
                log_proj[isinf(log_proj)]=0 
                d = (-proj.reshape(sinosize) + sinogram.reshape(sinosize) * log_proj.reshape(sinosize)).sum()
                print "func:",d
                return -float64(d)

    def get_gradient_activity(self,activity, activity_size, smoothing=0.0):
                """Returns the derivative of the log-likelihood with respect to the activity. At the moment, this 
                implements only the Poisson likelihood. """
                eps = 1e-16
                sinogram = self._measurement
                psf = self._psf
                attenuation = None
                sinoshape = sinogram.shape
                activity = activity.reshape(activity_size)
                if any(activity<0): 
                    activity[activity<0]=0
                proj = self.project(activity).data + eps       #FIXME: optionally use subsets 
                norm = self.get_normalization()  
                back = self.backproject(sinogram.reshape(sinoshape)/proj.reshape(sinoshape)).data+eps
                grad = (back-norm)
                #print "SIZE0: ",activity_size, proj.shape, norm.shape, grad.shape
                kernel = asarray([[0,1,0],[1,-4,1],[0,1,0]])
                a = activity.reshape([activity_size[0],activity_size[1],activity_size[2]])
                prior = asfortranarray(smoothing * scipy.signal.convolve2d(a,kernel,'same'))
                G = grad.reshape(activity.size) + prior.reshape(activity.size)
                G = G.reshape(activity_size[0]*activity_size[1]*activity_size[2])
                return -float64(G)

    def get_likelihood_log(self, activity_log, activity_size, smoothing=0.0): 
                """Returns the likelihood value - given the log of the activity. 
                This at the moment implements the Poisson likelihood only. This is useful to optimize the likelihood 
                or the posterior with respect to the log of the activity - e.g. to include the non negativity constraint. """
                return self.get_likelihood(exp(activity_log), activity_size, smoothing)

    def get_gradient_log_activity(self, activity_log, activity_size, smoothing=0.0):
                """Returns the derivative of the log-likelihood with respect to the log of the activity. At the moment, this 
                implements only the Poisson likelihood. This is useful to optimize the likelihood or the posterior with respect to the 
                log of the activity - e.g. to include the non negativity constraint. """
                activity = exp(activity_log)
                g = self.get_gradient_activity(activity, activity_size, smoothing)
                return float64(g*activity)

    def volume_render(self,volume,scale=1.0): 
        # FIXME: use the VolumeRenderer object in occiput.Visualization (improve it), the following is a quick fix: 
        if isinstance(volume,ndarray): 
            volume = float32(volume)
        else: 
            volume = float32(volume.data)
        proj = self.project(volume).data
        proj[where(proj>proj.max()/scale )]=proj.max()/scale
        return SPECT_Projection(proj)

    def load_measurement_file(self,filename): 
        pass 

    def set_measurement(self, measurement): 
        if not ( self._p_n_pix_x == measurement.shape[0] and self._p_n_pix_y == measurement.shape[1] and self._p_gantry_angular_positions == measurement.shape[2] ):
            raise UnexpectedParameter('Measurement size is not compatible with n_pix_x, n_pix_y, gantry_angular_positions. ')
        self._measurement = measurement 

    def load_measurement_from_file(self,measurement_file):
        measurement = import_nifti(measurement_file).data
        self.set_measurement(measurement)
        
    def get_measurement(self): 
        return Volume(self._measurement)

    def display_measurement(self): 
        return SPECT_Projection(self._measurement)

    def _make_svg(self): 
        if not has_svgwrite: 
            self._svg_string = None 
            return self._svg_string 

        w = '100%'
        h = '100%'
        
        dwg = svgwrite.Drawing('SPECT.svg',size=(w,h), profile='full', debug=True)
        dwg.viewbox(width=100, height=100)

        # DETECTOR 
        # collimator 
        rect = dwg.add(dwg.rect(insert=(12, 30), size=(8, 40), rx=0.5, ry=0.5))
        rect.fill('grey',opacity=0.5).stroke('black',width=0.3,opacity=0.001)

        # scintillator
        rect = dwg.add(dwg.rect(insert=(9, 30), size=(3, 40), rx=0.5, ry=0.5))
        rect.fill('green',opacity=0.1).stroke('none',width=0.3,opacity=0.001)

        # photomultipliers 
        for i in range(8): 
            rect = dwg.add(dwg.rect(insert=(1, 31.2+i*4.8), size=(8, 4), rx=0.3, ry=0.3))
            rect.fill('grey',opacity=0.25).stroke('none',width=0.3,opacity=0.001)
        
        # IMAGING VOLUME
        rect = dwg.add(dwg.rect(insert=(30, 30), size=(40, 40), rx=0.5, ry=0.5))
        rect.fill('grey',opacity=0.02).stroke('grey',width=0.3,opacity=0.02)
        
        # GEOMETRIC NOTATIONS 
        # circle, gantry rotation 
        circle = dwg.add(dwg.circle(center=(50, 50), r=30))
        circle.fill('none').stroke('grey', width=0.1).dasharray([0.5, 0.5]) 
        # center
        circle = dwg.add(dwg.circle(center=(50, 50), r=0.5))
        circle.fill('grey',opacity=0.1).stroke('grey', width=0.1)    
        line = dwg.add(dwg.line(start=(50-1,50), end=(50+1,50)))
        line.stroke('grey', width=0.1) 
        line = dwg.add(dwg.line(start=(50,50-1), end=(50,50+1)))
        line.stroke('grey', width=0.1) 
        
        #line = dwg.add(dwg.polyline([(10, 10), (10, 100), (100, 100), (100, 10), (10, 10)],stroke='black', fill='none'))
        self._svg_string = dwg.tostring() 
        return self._svg_string

    def _repr_svg_(self): 
        self._make_svg()
        return self._svg_string    





class Gantry(): 
    def __init__(self): 
        self.svg_string = self.make_svg()

    def make_svg(self):
        if not has_svgwrite: 
            self._svg_string = None 
            return self._svg_string 

        w = '100%'
        h = '100%'
        
        dwg = svgwrite.Drawing('test.svg',size=(w,h), profile='full', debug=True)
        dwg.viewbox(width=100, height=100)

        # DETECTOR 
        # collimator 
        rect = dwg.add(dwg.rect(insert=(12, 30), size=(8, 40), rx=0.5, ry=0.5))
        rect.fill('grey',opacity=0.5).stroke('black',width=0.3,opacity=0.001)

        # scintillator
        rect = dwg.add(dwg.rect(insert=(9, 30), size=(3, 40), rx=0.5, ry=0.5))
        rect.fill('green',opacity=0.1).stroke('none',width=0.3,opacity=0.001)

        # photomultipliers 
        for i in range(8): 
            rect = dwg.add(dwg.rect(insert=(1, 31.2+i*4.8), size=(8, 4), rx=0.3, ry=0.3))
            rect.fill('grey',opacity=0.25).stroke('none',width=0.3,opacity=0.001)
        
        # IMAGING VOLUME
        rect = dwg.add(dwg.rect(insert=(30, 30), size=(40, 40), rx=0.5, ry=0.5))
        rect.fill('grey',opacity=0.02).stroke('grey',width=0.3,opacity=0.02)
        
        # GEOMETRIC NOTATIONS 
        # circle, gantry rotation 
        circle = dwg.add(dwg.circle(center=(50, 50), r=30))
        circle.fill('none').stroke('grey', width=0.1).dasharray([0.5, 0.5]) 
        # center
        circle = dwg.add(dwg.circle(center=(50, 50), r=0.5))
        circle.fill('grey',opacity=0.1).stroke('grey', width=0.1)    
        line = dwg.add(dwg.line(start=(50-1,50), end=(50+1,50)))
        line.stroke('grey', width=0.1) 
        line = dwg.add(dwg.line(start=(50,50-1), end=(50,50+1)))
        line.stroke('grey', width=0.1) 
        
        #line = dwg.add(dwg.polyline([(10, 10), (10, 100), (100, 100), (100, 10), (10, 10)],stroke='black', fill='none'))
        return dwg.tostring() 

    def _repr_svg_(self): 
        return self.svg_string 




class GE_Infinia(SPECT_Static_Scan):
    def __init__(self): 
        SPECT.__init__(self)
        self._name = "GE Infinia SPECT Scanner with LEHR collimator"  




# occiput 
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# March 2015, Boston

__all__ = ['KSpace', 'MR_Static_Scan', 'MR_Dynamic_Scan'] 

# Import occiput: 
from occiput.Core import Image3D
from occiput.Core import Transform_Identity
from occiput.Visualization import ProgressBar, svgwrite, has_svgwrite, ipy_table, has_ipy_table
from occiput.Visualization.Colors import *
from occiput.DataSources.Synthetic.Shapes import uniform_cylinder
from occiput.Visualization import ipy_table, has_ipy_table, svgwrite, has_svgwrite 
from occiput.Core.NiftyPy_wrap import has_NiftyPy, INTERPOLATION_POINT, INTERPOLATION_LINEAR
from occiput.DataSources.FileSources.vNAV import load_vnav_mprage
#from occiput.DataSources.FileSources.MR_motion_sensors import load_motion_sensor_data
from occiput.DataSources.FileSources import import_nifti

# Import DisplayNode for IPython integration
from DisplayNode import DisplayNode 

# Import other modules
from PIL import ImageDraw
from PIL import Image as PIL 
from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, float32, where, ndarray, nan, inf, exp, asarray, complex64, complex128, complex, abs, angle, real, imag
from numpy.fft import fftn, ifftn, fft2, ifft2, fftshift, ifftshift
from numpy.random import randint 
from numpy.random import randint 
from numpy import asfortranarray, asfortranarray
from scipy import optimize 
import scipy
import scipy.signal
import scipy.io
import scipy.ndimage
import os
import h5py 

# Import ilang (inference language; optimisation) 
from MR_ilang import MR_Static_Gaussian, MR_Dynamic_Gaussian, ProbabilisticGraphicalModel
from ilang.Samplers import Sampler 

# Set verbose level
from occiput.global_settings import *
set_verbose_no_printing()
#set_verbose_high()


INTERPOLATION_COMPLEXPLANE = 0
INTERPOLATION_POLAR        = 1



class SequenceParameters(): 
    def __init__(self):
        self.Tr = 0.020
        self.name = 'Flash'



class KSpace: 
    def __init__(self, filename=None): 
        self.parameters = SequenceParameters() 
        self.data = None 
        self.motion = None 
        self.motion_events_indexes = None 
        self.filename = None 
        if filename is not None: 
            self.load_from_file(filename) 

    def load_from_file(self, filename, motion_data_file=None): 
        self.filename = filename 
        mat = scipy.io.loadmat(filename) 
        self.data = mat['data'] 
        # fftshift
        for i in range(self.data.shape[0]): 
            self.data[i,:,:,:] = fftshift(self.data[i,:,:,:])
        #self.data = self.data.squeeze() 
        if motion_data_file is not None: 
            self.load_motion_from_file(motion_data_file) 

    def load_motion_from_file(self, filename): 
        motion = load_motion_sensor_data(filename) 
        self.motion = motion      # FIXME: define motion information 
        motion_events = self.motion.extract_motion_events() 
        motion_events_indexes = where(motion_events!=0)[0].tolist() 
        self.motion_events_indexes = motion_events_indexes 

    def get_static_data(self): 
        data = complex64(zeros( (self.data.shape[1],self.data.shape[2],self.data.shape[3]) )) 
        for i in range(self.data.shape[0]): 
            data = data + self.data[i,:,:,:] 
        return data 


class MR_Static_Scan: 
    def __init__(self): 
        self.kspace = None 
    
    def load_kspace(self,filename): 
        self.kspace = KSpace() 
        self.kspace.load_from_file(filename) 
        self._construct_ilang_model() 
    
    def reconstruct_iterative(self, method=None, iterations=20, tol=None, dt=0.4, beta=0.05): 
        kspace = self.kspace.get_static_data()
        shape = kspace.shape
        image = asarray(complex64(zeros(shape)),order='F')
        for iter in range(iterations): 
            diff = kspace - self.project(image)
            grad = -self.backproject(diff) 
            image_abs = abs(image) 
            kernel = -ones((3,3,3))
            kernel[1,1,1]=26; 
            # 1) smooth magnitude 
            eps = 1e-8
            norm = 2*(scipy.ndimage.filters.convolve(image_abs,kernel)+eps) / (image_abs+eps);  # !!!
            # 2) smooth square magnitude 
            #norm = 4*(scipy.ndimage.filters.convolve(image_abs,kernel))              
            g0_real = -real(image) * norm
            g0_imag = -imag(image) * norm
            g0 = complex64(g0_real+g0_imag*1j)
            image = image - dt * (grad - beta*g0) 
        return Image3D(abs(image)) 

    def reconstruct_ifft(self): 
        kspace = self.kspace.get_static_data() 
        image = asarray( fftshift(ifftn(kspace)), order='F')
        return Image3D(abs(image)) 

    def project(self, volume ): 
        return asarray( fftn(fftshift(volume)), order='F')

    def backproject(self, kspace ): 
        return asarray( fftshift(ifftn(kspace)), order='F' )

    def _construct_ilang_model(self): 
        # define the ilang probabilistic model 
        self.ilang_model = MR_Static_Gaussian(self) 
        # construct a basic Directed Acyclical Graph 
        self.ilang_graph = ProbabilisticGraphicalModel(['x','k','sigma']) 
        self.ilang_graph.set_nodes_given(['k','sigma'],True) 
        self.ilang_graph.add_dependence(self.ilang_model,{'x':'x','k':'k','sigma':'sigma'}) 
        # construct a basic sampler object 
        #self.sampler     = Sampler(self.ilang_graph) 

    def _repr_html_(self): 
        if not has_ipy_table: 
            return "Please install ipy_table." 
        table_data = [ 
        ['bla   ',1],
        ['blabla',2], ]

        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic_left')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        return table._repr_html_() 

  

class MR_Dynamic_Scan: 
    def __init__(self): 
        self.kspace = None
        self.time_bins = None
    
    def load_kspace(self,filename, motion_data_file=None):
        self.kspace = KSpace() 
        self.kspace.load_from_file(filename, motion_data_file)
        N_time_bins = len(self.kspace.motion_events_indexes) + 1
        self.time_bins = [0]+self.kspace.motion_events_indexes
        self._construct_ilang_model()

    def reconstruct_iterative(self, method=None,  iterations=20, tol=None, dt=0.9, beta=0.05, active_frames=None, interpolation_space=INTERPOLATION_COMPLEXPLANE ,interpolation_mode=INTERPOLATION_POINT):  
        shape = self.kspace.get_static_data().shape
        image = asarray(complex128(zeros(shape)),order='F')
        n_frames = len(self.time_bins) 
        for iter in range(iterations): 
            grad = asarray(complex128(zeros(shape)),order='F') 
            for framei in range(n_frames): 
                use_frame=True 
                if active_frames is not None: 
                    if framei in active_frames: 
                        use_frame=True 
                    else: 
                        use_frame=False 
                if use_frame: 
                    print "time frame: %d / %d "%(framei,n_frames) 
                    if framei==0: 
                        image_tr = image 
                    else: 
                        M = self._affine 
                        image_tr = self._transform_image(image,M,interpolation_space=interpolation_space, interpolation_mode=interpolation_mode)
                    kspace = asarray(self.kspace.data[framei,:,:,:],order='F') 
                    diff = kspace - self.project(image_tr) 
                    grad_tr = -self.backproject(diff) 
                    if framei==0: 
                        grad += grad_tr 
                    else: 
                        grad += self._transform_image(grad_tr,M.inverse, interpolation_space=interpolation_space, interpolation_mode=interpolation_mode)
            image_abs = abs(image) 
            kernel = -ones((3,3,3))
            kernel[1,1,1]=26; 
            # 1) smooth magnitude 
            eps = 1e-8
            norm = 2*(scipy.ndimage.filters.convolve(image_abs,kernel)+eps) / (image_abs+eps);  
            # 2) smooth square magnitude 
            #norm = 4*(scipy.ndimage.filters.convolve(image_abs,kernel))              
            g0_real = real(image) * norm
            g0_imag = imag(image) * norm
            g0 = g0_real + g0_imag*1j
            image = image - dt * (grad + beta*g0) 
        return Image3D(abs(image)) 

    def _set_affine(self,affine): 
        self._affine = affine

    def _transform_image(self, image, affine, interpolation_space=INTERPOLATION_COMPLEXPLANE ,interpolation_mode=INTERPOLATION_POINT): 
        I  = Image3D(zeros(image.shape)) 
        I.set_space('world') 
        grid = I.get_world_grid() 
        I.transform(affine) 
        if interpolation_space == INTERPOLATION_COMPLEXPLANE: 
            re = asarray(real(image),order='F') 
            im = asarray(imag(image),order='F') 
            I.data = re
            re = I.compute_resample_on_grid(grid,interpolation_mode=interpolation_mode).data
            I.data = im
            im = I.compute_resample_on_grid(grid,interpolation_mode=interpolation_mode).data 
            image_tr = complex128(re+im*1j)
        elif interpolation_space == INTERPOLATION_POLAR: 
            r = asarray(abs(image),order='F')
            t = asarray(angle(image),order='F')
            I.data = r
            r = I.compute_resample_on_grid(grid,interpolation_mode=interpolation_mode).data
            I.data = t
            t = I.compute_resample_on_grid(grid,interpolation_mode=interpolation_mode).data  
            image_tr = r * exp(t*1j)
        return image_tr
        
    def reconstruct_ifft(self): 
        kspace = self.kspace.get_static_data() 
        image = asarray( fftshift(ifftn(kspace)), order='F' )
        return Image3D(abs(image))

    def project(self, volume ): 
        return asarray( fftn(fftshift(volume)), order='F')

    def backproject(self, kspace ): 
        return asarray( fftshift(ifftn(kspace)), order='F' )

        
    def _construct_ilang_model(self): 
        # define the ilang probabilistic model 
        self.ilang_model = MR_Dynamic_Gaussian(self)  

    def _repr_html_(self): 
        if not has_ipy_table: 
            return "Please install ipy_table."
        table_data = [ 
        ['bla   ',1],
        ['blabla',2], ]

        table = ipy_table.make_table(table_data)
        table = ipy_table.apply_theme('basic_left')
        table = ipy_table.set_global_style(float_format="%3.3f")        
        return table._repr_html_() 






# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# Apr. 2014, Boston, MA


from ilang_models import SSD_ilang

from registration import Registration_Two_Images, Registration_Longitudinal, Registration_N_Images






class BaseScintillatorSPECT(): 
    def __init__(self): 
        self._name = "Generic Scintillator"
        self._manufacturer = "No manufacturer"
        self._scintillator_material = "Not specified"
        self._p_thickness_mm = 10.0
        self._need_to_recompute_psf = True
        self._previous_energy_request = 0

    def get_name(self): 
        return self._name

    def get_manufacturer(self): 
        return self._manufacturer

    def get_material(self): 
        return self._scintillator_material

    def get_parameters(self):
        parameters = {}
        dic = self.__dict__
        for k in dic.keys(): 
            if k.startswith('_p_'):
                parameters[k[3:]]=dic[k]        
        return parameters
 
    def get_psf(self,energy_kev): 
        if self._need_to_recompute_psf or energy_kev!=self._previous_energy_request: 
            self._compute_psf(energy_kev)
            self._previous_energy_request = energy_kev
        return self._psf 

    def _compute_psf(self,energy_kev): 
        print_medium_verbose("Recomputing PSF..")
        self._need_to_recompute_psf = False
        self._psf = 0


class Ideal(BaseScintillatorSPECT): 
    def __init__(self): 
        BaseScintillatorSPECT.__init__(self)
        self._name = "Ideal Scintillator"

    def get_psf(self,energy_kev): 
        if self._need_to_recompute_psf or energy_kev!=self._previous_energy_request: 
            self._compute_psf(energy_kev)
            self._previous_energy_request = energy_kev
        return self._psf 

    def _compute_psf(self,energy_kev): 
        print_medium_verbose("Recomputing PSF scintillator..")
        self._need_to_recompute_psf = False
        self._psf = 0








# occiput.io 
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# 2013 - 2015
# Boston, MA


# Abstraction of a PET scanner: probabilistic graphical model


import ilang 
import ilang.Models 
from ilang.Models import Model 
from ilang.Graphs import ProbabilisticGraphicalModel 

__all__ = ['PET_Static_Poisson','PET_Dynamic_Poisson','ProbabilisticGraphicalModel']



class PET_Static_Poisson(Model): 
    variables = {'lambda':'continuous','alpha':'continuous','counts':'discrete'} 
    dependencies = [['lambda','counts','directed'],['alpha','counts','directed']]

    def __init__(self, PET_scan, name=None):
        if name == None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
        # PET scan
        self.PET_scan = PET_scan    
        # small number
        self.EPS = 1e9

    def set_PET_scan(self, PET_scan): 
        self.PET_scan = PET_scan 

    def init(self): 
        pass 

    def log_conditional_probability_lambda(self,lambda_): 
        alpha = self.get_value('alpha')
        return 0 

    def log_conditional_probability_gradient_lambda(self,lambda_): 
        alpha = self.get_value('alpha')
        projection_data = self.PET_scan.project(lambda_, alpha) 
        Nx = lambda_.shape[0]; Ny = lambda_.shape[1]; Nz = lambda_.shape[2]
        gradient        = self.PET_scan.backproject(self.PET_scan.get_measurement()[0]/(projection_data[0]+self.EPS), Nx,Ny,Nz, alpha) 
        return gradient 

    def log_conditional_probability_alpha(self,alpha): 
        lambda_ = self.get_value('lambda') 
        return 0 

    def log_conditional_probability_gradient_alpha(self,alpha): 
        lambda_ = self.get_value('lambda')  
        return 0
        
    def sample_conditional_probability_z(self): 
        return 0    



class PET_Dynamic_Poisson(Model): 
    variables = {'lambda':'continuous','alpha':'continuous','roi_1':'continuous','roi_1':'continuous','counts_1':'discrete','counts_2':'discrete'} 
    dependencies = [['lambda','counts_1','directed'],['lambda','counts_2','directed'],['alpha','counts_1','directed'],['alpha','counts_2','directed'],['roi_1','counts_1','directed'],['roi_2','counts_2','directed']] 

    def __init__(self, PET_scan, name=None): 
        if name == None:  
            name = self.__class__.__name__
        Model.__init__(self, name) 
        # PET scan object: 
        self.PET_scan = PET_scan 
        self.N_time_bins = len(self.PET_scan.time_bins)
        # Variables and dependencies: 
        self._lambda = None 
        self._alpha  = None 
        self._rois   = None 
        self.variables = {'lambda':'continuous', 'alpha':'continuous'} 
        self.dependencies = []
        for t in range(self.N_time_bins): 
            var_name_counts = 'counts_'+str(t)
            var_name_roi    = 'roi_'+str(t)
            self.variables[var_name_counts]='discrete'
            self.variables[var_name_roi]='continuous'
            self.dependencies.append(['lambda',var_name_counts,'directed'])
            self.dependencies.append(['alpha',var_name_counts,'directed']) 
            self.dependencies.append([var_name_roi,var_name_counts,'directed']) 

    def set_PET_scan(self, PET_scan): 
        self.PET_scan = PET_scan 

    def init(self): 
        pass 

    def log_conditional_probability_lambda(self): 
        return 0 

    def log_conditional_probability_gradient_lambda(self): 
        return numpy.zeros([100,1])

    def log_conditional_probability_alpha(self): 
        return 0 

    def log_conditional_probability_gradient_alpha(self): 
        return numpy.zeros([100,1])
        
    def sample_conditional_probability_counts(self): 
        return 0    





        



# occiput 
# March 2015 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA 


__all__ = ['load_motion_sensor_data'] 

from occiput.Core import Transform_Affine 
import numpy 
from numpy import float32, int32 
import matplotlib.pyplot as plt
import copy 
from occiput.Core import transformations as tr

BOX_MIN      = [-50.0,-50.0,-50.0]
BOX_MAX      = [50.0,50.0,50.0]
THRESHOLD_MM = 5.0
LINE_COLOR   = "#2f8dff"


def quaternion_to_rotation(q,axes='sxyz'): 
    return tr.euler_from_quaternion(q,axes) 

def angle_axis_to_quaternion(angle_rad,axis):
    f = numpy.sin(0.5*angle_rad)
    quaternion = numpy.asarray( [numpy.cos(0.5*angle_rad),f*axis[0],f*axis[1],f*axis[2]] )
    return quaternion

def angle_axis_to_rotation(angle_rad,axis): 
    quaternion = angle_axis_to_quaternion(angle_rad,axis)
    rotation = quaternion_to_rotation(quaternion) 
    return rotation

def affine_from_quaternion(q,axes='sxyz'): 
    pass 

def quaternion_from_matrix(affine):
    return tx.quaternion_from_matrix(affine)

def rad_to_deg(rad): 
    return numpy.asarray(rad)*180.0/numpy.pi
    
def deg_to_rad(deg): 
    return numpy.asarray(deg)/180.0*numpy.pi





class Motion_Sensor: 
    def __init__(self,filename=None, channels=['B']): 
        self._reset()
        if filename is not None: 
            self.load_from_file(filename, channels) 

    def get_motion_quaternion(self, index): 
        motion_affine = self.get_motion_affine(index) 
        return quaternion_from_matrix(motion_affine)

    def get_motion_affine(self, index): 
        return self._motion[index] 

    def _reset(self):
        self._motion = []
        self._n_time_points = 0 
        self._tx = []; self._ty = []; self._tz = []; 
        self._rx = []; self._ry = []; self._rz = []; 
        self._q0 = []; self._q1 = []; self._q2 = []; self._q3 = [];

    def get_n_time_points(self):
        return self._n_time_points

    def load_from_file(self,filename, channels=['B']): 
        f = open(filename)
        self._reset() 
        while 1: 
            l = f.readline() 
            if l=='': 
                break 
            d = l.split() 
            channel = d[0][1]
            
            do_process = False 
            if channels is None: 
                do_process = True
            elif channels == []: 
                do_process = True 
            elif channels == '': 
                do_process = True 
            else: 
                if channel in channels: 
                    do_process = True
                else: 
                    do_process = False

            if do_process: 
                data    = d[1]
                x  = float32(d[2])
                y  = float32(d[3])
                z  = float32(d[4])
                q0 = float32(d[5])
                q1 = float32(d[6])
                q2 = float32(d[7])
                q3 = float32(d[8])
                q = [q0,q1,q2,q3]
                #q = tr.quaternion_conjugate(q)
                status = int32(d[11])
                uncertainty = float32(d[12]) 
                self._n_time_points += 1
                tra_mat = tr.translation_matrix([x,y,z]) 
                rot_mat = tr.quaternion_matrix(q)
                self._motion.append( numpy.dot(tra_mat,rot_mat) )
                rotation = quaternion_to_rotation(q)
                self._tx.append(x)
                self._ty.append(y)
                self._tz.append(z)
                self._rx.append(rotation[0])
                self._ry.append(rotation[1])
                self._rz.append(rotation[2])
                self._q0.append(q[0])
                self._q1.append(q[1])
                self._q2.append(q[2])
                self._q3.append(q[3])

    def _draw_rectangle(self, axis, x, y, alpha=0.2, ec="gray", fc="gray"): #"CornflowerBlue"
        axis.add_patch( plt.Rectangle( (x[0],y[0]) , x[1]-x[0],y[1]-y[0], alpha=alpha, ec=ec, fc=fc, visible=True) )
        #plt.draw() 

    def _draw_rectangles(self, axis, windows, range_y): 
        for ii in range(len(windows)):
            tt = windows[ii]
            yy = (range_y[0],range_y[1])
            self._draw_rectangle(axis,tt,yy)

    def _draw_line(self, axis, x, range_y, color="#ff8d8d", linestyle="dashed", label=""): 
        axis.vlines(x, range_y[0], range_y[1], colors=color, linestyles=linestyle, label=label, visible=True)

    def _draw_events(self, axis, time, range_y, events, color="#ff8d8d", linestyle="dashed", label=""): 
        for t_index in time: 
            if t_index: 
              if t_index<self.get_n_time_points():
                if events[t_index-1]: 
                    #print "Drawing line: ", t_index, range_y, color, linestyle
                    self._draw_line(axis, t_index, range_y, color, linestyle, label)

    def plot_motion(self, save_to_file=None, extract_events_threshold=THRESHOLD_MM, method='box', box_min=BOX_MIN, box_max=BOX_MAX, min_duration=10, line_color=LINE_COLOR):
        t = range(len(self._tx))

        # make windows:
        if extract_events_threshold is not None: 
            windows = []
            events  = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)
        t_index_start = 0
        for t_index in t:  
            if t_index:     #this excludes the possibility of a motion event at time 0 
              if t_index<self.get_n_time_points():
                if events[t_index-1]: 
                    t_index_end = t_index-1
                    if t_index_end-t_index_start > min_duration: 
                        windows.append( (t_index_start, t_index_end) )  #end window with frame before a motion event 
                    t_index_start = t_index+1                 #start window with frame after a motion event 
        windows.append( (t_index_start, t[-1]) ) 

        if 1:      
            fig1 = plt.figure(1,figsize=(17.5,4),dpi=200)

            ax1 = fig1.add_subplot(321)
            ax1.plot(t, self._tx, line_color)
            ax1.grid(True)
            pr0 = float32(self._tx).min()
            pr1 = float32(self._tx).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('TX [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 

            ax1 = fig1.add_subplot(323)
            ax1.plot(t, self._ty, line_color)
            ax1.grid(True)
            pr0 = float32(self._ty).min()
            pr1 = float32(self._ty).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('TY [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 
    
            ax1 = fig1.add_subplot(325)
            ax1.plot(t, self._tz, line_color)
            ax1.grid(True)
            pr0 = float32(self._tz).min()
            pr1 = float32(self._tz).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('TZ [mm]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 

#            if save_to_file is not None: 
#                plt.savefig(save_to_file)

            fig2 = fig1 #plt.figure(2)
    
            ax1 = fig2.add_subplot(322)
            ax1.plot(t, rad_to_deg(self._rx), line_color)
            ax1.grid(True)
            pr0 = float32(rad_to_deg(self._rx)).min()
            pr1 = float32(rad_to_deg(self._rx)).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('RX [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 
            
            ax1 = fig2.add_subplot(324)
            ax1.plot(t, rad_to_deg(self._ry), line_color)
            ax1.grid(True)
            pr0 = float32(rad_to_deg(self._ry)).min()
            pr1 = float32(rad_to_deg(self._ry)).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('RY [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 
                                
            ax1 = fig2.add_subplot(326)
            ax1.plot(t, rad_to_deg(self._rz), line_color)
            ax1.grid(True)
            pr0 = float32(rad_to_deg(self._rz)).min()
            pr1 = float32(rad_to_deg(self._rz)).max()
            d = pr1-pr0
            pr = [pr0-d/4.0, pr1+d/4.0]
            ax1.set_ylim( pr )
            ax1.set_ylabel('RZ [deg]')
            #for label in ax1.get_xticklabels():
            #    label.set_color('r')
            if extract_events_threshold  is not None:
                self._draw_rectangles(ax1,windows,pr) 
                self._draw_events(ax1, t, pr, events) 
                                
#            if save_to_file is not None: 
#                plt.savefig(save_to_file)
            
        plt.show()
        #return fig1,fig2


    def get_mean_displacement(self, index, method='box', box_min=BOX_MIN, box_max=BOX_MAX): 
        mat = self.get_motion_affine(index)
        mat = Transform_Affine(mat) 
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners[3,:]  =0
            #print "CORNERS: ", corners_t
            #dist = numpy.sqrt(((corners-corners_t)**2).sum(0))
            dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ
 
    def get_mean_displacement_variation(self, index, method='box', box_min=BOX_MIN, box_max=BOX_MAX):
        mat  = self.get_motion_affine(index)
        if index > 0: 
            mat0 = self.get_motion_affine(index-1)
        else: 
            mat0 = tr.identity_matrix()
        mat = Transform_Affine(mat) 
        mat0 = Transform_Affine(mat0)
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners_t0 = mat0.left_multiply(corners)
            corners_t0[3,:]=0
            dist = numpy.sqrt(((corners_t-corners_t0)**2).sum(0))
            #dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ

    def get_mean_displacement_variation_since_time(self, index_new, index_old, method='box', box_min=BOX_MIN, box_max=BOX_MAX):
        mat  = self.get_motion_affine(index_new)
        mat0 = self.get_motion_affine(index_old)
        mat = Transform_Affine(mat) 
        mat0 = Transform_Affine(mat0)
        if method=='box': 
            b = box_min
            B = box_max
            corners = numpy.asarray([ [b[0],b[1],b[2],1], [B[0],b[1],b[2],1], [b[0],B[1],b[2],1], [B[0],B[1],b[2],1], [b[0],b[1],B[2],1], [B[0],b[1],B[2],1], [b[0],B[1],B[2],1], [B[0],B[1],B[2],1] ]).transpose()
            corners_t = mat.left_multiply(corners)
            corners_t[3,:]=0
            corners_t0 = mat0.left_multiply(corners)
            corners_t0[3,:]=0
            dist = numpy.sqrt(((corners_t-corners_t0)**2).sum(0))
            #dist = (corners-corners_t).sum(0)
            mean_displ = numpy.mean(dist)
        else: 
            raise "Method to compute mean displacement is unknown. "
        return mean_displ

        
    def extract_motion_events(self, method='box', threshold=THRESHOLD_MM, box_min=BOX_MIN, box_max=BOX_MAX, prune_distance=50):
        t = range(self.get_n_time_points())
        is_event   = numpy.zeros(len(t)-1)
        t_index_old = 0
        for t_index in t[1:]: 
#             mean_displ = self.get_mean_displacement_variation(t_index, method, box_min, box_max ) 
#             if numpy.sqrt((mean_displ)**2) >= threshold: 

             mean_displ = self.get_mean_displacement_variation_since_time(t_index, t_index_old, method, box_min, box_max )
             if numpy.sqrt((mean_displ)**2) >= threshold: 
                 t_index_old = numpy.copy(t_index)
                 is_event[t_index-1] = 1
             else: 
                 is_event[t_index-1] = 0 
        if prune_distance >=2: 
            last_event = 0
            for i in range(len(is_event)): 
                if is_event[i]: 
                    if (i-last_event) <= prune_distance:
                        is_event[i]=0
                    else:     
                        last_event = i
        return is_event 


    def plot_mean_displacement(self, method='box', box_min=BOX_MIN,box_max=BOX_MAX, save_to_file=None, plot_zero=False, extract_events_threshold=THRESHOLD_MM, plot_range=[None,None], line_color=LINE_COLOR, min_duration=10 ): 
        t = range(self.get_n_time_points())
        mean_displ = numpy.zeros(len(t))
        mean_displ_var = numpy.zeros(len(t))
        mean_displ_var_since_event = numpy.zeros(len(t))
        
        if extract_events_threshold is not None: 
            events = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)

        t_index_old = 0
        for t_index in t: 
             mean_displ[t_index] = self.get_mean_displacement(t_index, method, box_min, box_max ) 
             mean_displ_var[t_index] = self.get_mean_displacement_variation(t_index, method, box_min, box_max ) 
             mean_displ_var_since_event[t_index] = self.get_mean_displacement_variation_since_time(t_index, t_index_old, method, box_min, box_max ) 
             if t_index: 
                 if events[t_index-1] == 1: 
                     t_index_old = t_index-1
        if not plot_zero: 
            t = t[1:]
            mean_displ = mean_displ[1:]
            mean_displ_var = mean_displ_var[1:]
            mean_displ_var_since_event = mean_displ_var_since_event[1:]

        # make windows:
        if extract_events_threshold is not None: 
            windows = []
            events  = self.extract_motion_events(method, extract_events_threshold, box_min, box_max)
        t_index_start = 0
        for t_index in t:  
            if t_index:     #this excludes the possibility of a motion event at time 0 
                if events[t_index-1]: 
                    t_index_end = t_index-1
                    if t_index_end-t_index_start > min_duration: 
                        windows.append( (t_index_start, t_index_end) )  #end window with frame before a motion event 
                    t_index_start = t_index+1                 #start window with frame after a motion event 
        windows.append( (t_index_start, t[-1]) ) 
        
#        mean_displ[numpy.where(mean_displ==0)]=-1000
#        mean_displ_var[numpy.where(mean_displ_var==0)]=-1000
#        mean_displ_var_since_event[numpy.where(mean_displ_var_since_event==0)]=-1000
        
        fig = plt.figure(5,figsize=(8,4),dpi=200)

        ax1 = fig.add_subplot(311)
        ax1.set_title("Times frames - vNAV")
        ax1.plot(t,mean_displ, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('disp [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events) 


        ax1 = fig.add_subplot(312)
        #ax1.set_title("Mean displacement delta ")
        ax1.plot(t,mean_displ_var, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ_var.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ_var.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('delta [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ_var
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events)             

        ax1 = fig.add_subplot(313)
        #ax1.set_title("Mean displacement event")
        ax1.plot(t,mean_displ_var_since_event, line_color)
        ax1.grid(True)
        if plot_range[0] is None: 
            pr0=copy.copy(mean_displ_var_since_event.min())
        else: 
            pr0=copy.copy(plot_range[0])
        if plot_range[1] is None: 
            pr1=copy.copy(mean_displ_var_since_event.max())
        else: 
            pr1=copy.copy(plot_range[1])
        ax1.set_ylim( [pr0,pr1] )
        ax1.set_ylabel('event [mm]')
        if extract_events_threshold is not None: 
            ax1.hold(1)
            E = events*mean_displ_var_since_event
            E[numpy.where(E<0.1)]=-1000
            ax1.plot(t,E,'r.')
            #ax1.plot(t,0.5*(events*(mean_displ.max()-mean_displ.min())+mean_displ.min()),'r.')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
        self._draw_rectangles(ax1,windows,[pr0,pr1]) 
        self._draw_events(ax1, t, [pr0,pr1], events) 

        if save_to_file is not None: 
            plt.savefig(save_to_file)
        plt.show() 
        #return fig

    def plot_quaternion(self, save_to_file=None, line_color=LINE_COLOR): 
        t = range(self.get_n_time_points())[1:]
        s = rad_to_deg(numpy.asarray(self._q0))[1:]
        
        fig = plt.figure(6,figsize=(8,4),dpi=200)
        ax1 = fig.add_subplot(211)
        ax1.set_title("Rotation agnle [deg] vs. vNAV frame number")
        ax1.plot(t,s, line_color)
        ax1.grid(True)
        pr0 = s.min()
        pr1 = s.max() 
        d = pr1-pr0
        pr = [pr0-d/4.0, pr1+d/4.0]
        ax1.set_ylim( pr )
        ax1.set_ylabel('Rotation angle [deg]')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')

        arc = numpy.zeros(self.get_n_time_points()) 
        v0  = numpy.asarray( self._q1[0],self._q2[0],self._q3[0] ) 
        for t in range(self.get_n_time_points()): 
            vt  = numpy.asarray( self._q1[t],self._q2[t],self._q3[t] ) 
            arc[t] = numpy.dot( numpy.transpose(v0), vt ) 
        ax1 = fig.add_subplot(212) 
        ax1.set_title("Arc vs. vNAV frame number") 
        t = range(self.get_n_time_points()) 
        ax1.plot(t,arc, line_color)
        ax1.grid(True)
        pr0 = arc.min()
        pr1 = arc.max() 
        d = pr1-pr0
        pr = [pr0-d/4.0, pr1+d/4.0]
        ax1.set_ylim( pr )
        ax1.set_ylabel('Arc [steradians]')
        #for label in ax1.get_xticklabels():
        #    label.set_color('r')
    
        if save_to_file is not None: 
            plt.savefig(save_to_file)
        plt.show() 
        #return fig

    def _repr_html_(self): 
        self.plot_mean_displacement() 



def load_motion_sensor_data(filename, channels=['B',]): 
    return Motion_Sensor(filename, channels)
    






# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Dec. 2013, Boston
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Jan. 2015, Boston
# Feb. 2015, Helsinki
# Nov. 2015, Boston 


# Here is the library of PET scanners! Access to listmode data is provided by an external Package. 
# Don't be put off, the mechanism is quite simple. 

__all__ = ["Generic","Brain_PET","Biograph_mMR","Discovery_RX","get_scanner_by_name"]


from occiput.Reconstruction.PET.PET_meshing import Michelogram
from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose 


# Import scanner definitions 
try: 
    from Occiput_Interface_Brain_PET import Brain_PET
except: 
    Brain_PET = None
try: 
    from Occiput_Interface_Biograph_mMR import Biograph_mMR
except: 
    Biograph_mMR = None
try: 
    from Occiput_Interface_Discovery_RX import Discovery_RX
except: 
    Discovery_RX = None



class Generic(): 
    def __init__(self):
        self.model        = "Generic PET Scanner" 
        self.manufacturer = "Occiput's immagination"
        self.version      = "1.0"
        self.supports_listmode = False
        self.uses_meshing      = False

        self.N_u    = 128
        self.N_v    = 128
        self.size_u = 2.0*128 
        self.size_v = 2.0*128
        self.N_azimuthal                        = 5
        self.N_axial                            = 120 
        self.angles_azimuthal                   = float32([-0.5, -0.25, 0.0, 0.25, 0.5])
        self.angles_axial                       = float32( linspace(0,pi-pi/self.N_axial,self.N_axial) )

        self.scale_activity                     = 1.0

        self.activity_N_samples_projection_DEFAULT       = 150 
        self.activity_N_samples_backprojection_DEFAULT   = 150 
        self.activity_sample_step_projection_DEFAULT     = 2.0 
        self.activity_sample_step_backprojection_DEFAULT = 2.0 
        self.activity_shape_DEFAULT                      = [128,128,128] 
        self.activity_size_DEFAULT                       = float32([2.0, 2.0, 2.0])*float32(self.activity_shape_DEFAULT)

        self.attenuation_N_samples_projection_DEFAULT       = 150 
        self.attenuation_N_samples_backprojection_DEFAULT   = 150 
        self.attenuation_sample_step_projection_DEFAULT     = 2.0 
        self.attenuation_sample_step_backprojection_DEFAULT = 2.0 
        self.attenuation_shape_DEFAULT                      = [128,128,128] 
        self.attenuation_size_DEFAULT                       = float32([2.0, 2.0, 2.0])*float32(self.attenuation_shape_DEFAULT)

        self.listmode = None
        self.physiology = None 


def get_scanner_by_name(name): 
    if name == "Generic": 
        return Generic
    elif name=="Brain_PET" or name=="BrainPET": 
        return Brain_PET
    elif name=="Biograph_mMR" or name=="Siemens_Biograph_mMR" or name=="mMR" or name=="Siemens_mMR": 
        return Biograph_mMR
    elif name=="Discovery_RX" or name=="GE_Discovery_RX" or name=="GE_RX" or name =="RX": 
        return Discovery_RX
    else: 
        return None 


    

    




# occiput - Computational engine for volumetric imaging 
# Stefano Pedemonte
# Aalto University 
# Harvard University, Martinos Center for Biomedical Imaging 
# 2012 - 2014, Helsinki, FI
# 2013 - 2017, Boston, MA, USA


# Use old Python build system, otherwise the extension libraries cannot be found. FIXME 
import sys
for arg in sys.argv: 
    if arg=="install":
        sys.argv.append('--old-and-unmanageable') 

from setuptools import setup, Extension
from glob import glob 


setup(
    name='occiput',
    version='1.0.12',
    author='Stefano Pedemonte',
    author_email='stefano.pedemonte@gmail.com',
    packages=['occiput', 
              'occiput.test', 
              'occiput.notebooks',
              'occiput.Core',
              'occiput.Reconstruction', 
              'occiput.Reconstruction.PET', 
              'occiput.Reconstruction.SPECT', 
              'occiput.Reconstruction.CT', 
              'occiput.Reconstruction.MR',
              'occiput.Transformation', 
              'occiput.Registration',
              'occiput.Registration.Affine',
              'occiput.Registration.TranslationRotation',
              'occiput.Classification', 
              'occiput.DataSources', 
              'occiput.DataSources.Synthetic', 
              'occiput.DataSources.FileSources', 
              'occiput.Visualization',
              ],
    package_data={'occiput': ['Data/*.pdf','Data/*.png','Data/*.jpg','Data/*.svg','Data/*.nii','Data/*.dcm','Data/*.h5','Data/*.txt','Data/*.dat']},
    scripts=[],
    url='http://www.occiput.io/',
    license='LICENSE.txt',
    description='Tomographic Vision - PET, SPECT, CT, MRI reconstruction and processing.',
    long_description=open('README.rst').read(),
    keywords = ["PET", "SPECT", "emission tomography", "transmission tomography", "tomographic reconstruction","nuclear magnetic resonance"],
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Bio-Informatics" ],
    install_requires=[
        "numpy >= 1.12.0", 
        "matplotlib >= 1.4.0",
        "simplewrap >= 0.3.0",
        "NiftyPy >= 0.4.0",
        "petlink >= 0.3.0",  
        "interfile >= 0.3.0", 
        "DisplayNode >= 0.3.0", 
        "ilang >= 0.3.0", 
        "ipy_table >= 1.11.0",
        "nibabel >= 2.0.0",  
        "pydicom >= 0.9.0",
        "nipy >= 0.3.0",
        "jupyter >= 1.0.0",
        "h5py >= 2.3.0",
        "scipy >= 0.14.0",
        "pillow >= 2.8.0",
        "svgwrite >= 1.1.0" ] 
) 






# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2014, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA
# April 2014


def import_listmode(filename):
    return None


def convert_listmode_dicom_to_interfile(dicom_file, interfile_file_name_base): 
    print "Converting listmode dicom file to listmode interfile. "

    import dicom 
    d = dicom.read_file(dicom_file)
    
    header = d.get(('0029', '1010'))
    data   = d.get(('7fe1', '1010'))
    
    f = open(interfile_file_name_base+".l.hdr","w")
    f.write(header.value)
    f.close() 
    
    f = open(interfile_file_name_base+".l","wb")
    f.write(data.value)
    f.close() 
    print "Done. "



# occiput  
# Stefano Pedemonte
# Harvard University, Martinos Center for Biomedical Imaging 
# April. 2015, Boston, MA



class FileNotFound(Exception): 
    def __init__(self,msg,filename): 
        self.msg = str(msg) 
        self.filename = str(filename)
    def __str__(self): 
        return "Cannot find file '%s' (%s)."%(self.filename, self.msg)

class UnknownParameter(Exception): 
    def __init__(self,msg): 
        self.msg = str(msg) 
    def __str__(self): 
        return "Unkwnown parameter: %s"%(self.msg)

class UnexpectedParameter(Exception): 
    def __init__(self,msg): 
        self.msg = str(msg) 
    def __str__(self): 
        return "Unexpected parameter: %s"%(self.msg)







# occiput.io 
# Stefano Pedemonte
# Martinos Center for Biomedical Imaging, Harvard University/MGH, Boston
# Aalto University, Finland 
# 2013-2015


# Code to generate subsets for OSEM (ordered subsets expectation maximization) and algorithms of that sort, such 
# as stochastic optimization. 


__all__ = ["SubsetGenerator"]

from numpy import isscalar, linspace, int32, uint32, ones, zeros, pi, sqrt, float32, float64, where, ndarray, nan
from numpy import inf, asarray, concatenate, fromfile, maximum, exp, asfortranarray, fliplr, transpose 
from numpy.random import randint 



class SubsetGenerator():  
    """This object has the machinery to generate subsets of the axial and longitudinal directions for PET reconstruction (e.g. OSEM algorithm). 
    It implements various algorithms to create the subsets. """
    def __init__(self,N_azimuthal,N_axial):
        self._N_axial     = N_axial
        self._N_azimu     = N_azimuthal
        self._index = 0

    def new_subset(self, mode, subset_size, azimuthal_range=None): 
        """Returns a new subset. """
        if mode=='random': 
            return self._subsets_random_no_replacement(subset_size,azimuthal_range) 
        elif mode=='random_axial': 
            return self._subsets_random_axial(subset_size)
        elif mode=='ordered_axial': 
            return self._subsets_ordered_axial(subset_size)
        else: 
            raise UnexpectedParameter("'mode' parameter %s not recognised."%str(mode))
        
    def all_active(self): 
        """Returns full set.  """
        return ones((self._N_azimu, self._N_axial),dtype=uint32,order="C") 

    def _subsets_random_no_replacement(self,subset_size,azimuthal_range=None): 
        """Generates subsets randomly - no replacement. """
        if subset_size is None: 
            return self.all_active() 
        if subset_size>=self._N_axial*self._N_azimu: 
            return self.all_active() 
        M = zeros((self._N_azimu, self._N_axial),dtype=uint32,order="C") 
        n = 0
        while n<subset_size:
            active_axial = randint(self._N_axial)
            if azimuthal_range is None: 
                active_azimu = randint(self._N_azimu) 
            else: 
                active_azimu = azimuthal_range[ randint(len(azimuthal_range)) ]
            if M[active_azimu,active_axial] == 0: 
                M[active_azimu,active_axial] = 1
                n+=1
        return M

    def _subsets_ordered_axial(self, subset_size): 
        """Generates ordered subsets; use all azimuthal angles, subsample axially. """
        if subset_size is None: 
            return self.all_active() 
        if subset_size>=self._N_axial*self._N_azimu: 
            return self.all_active() 
        M = zeros((self._N_azimu, self._N_axial),dtype=uint32,order="C") 
        for i in range(self._index, self._N_axial, self._N_axial/subset_size): 
            M[:,i] = 1
        self._index += 1 
        if self._index == self._N_axial: 
            self._index = 0 
        return M
        
    def _subsets_random_axial(self, subset_size): 
        """Generates random subsets; use all azimuthal angles, subsample axially. """
        if subset_size is None: 
            return self.all_active() 
        if subset_size>=self._N_axial*self._N_azimu: 
            return self.all_active() 
        M = zeros((self._N_azimu, self._N_axial),dtype=uint32,order="C") 
        n = 0
        while n < subset_size:
            active_axial = randint(self._N_axial)
            if M[0,active_axial] == 0: 
                M[:,active_axial] = 1
                n+=1
        return M

    




# occiput 
# Stefano Pedemonte
# Aalto University, School of Science, Helsinki
# Oct 2013, Helsinki 
# Harvard University, Martinos Center for Biomedical Imaging 
# Boston, MA, USA



import nibabel
from occiput.Core.Conversion import nipy_to_occiput, occiput_from_array, nifti_to_occiput
from occiput.DataSources.FileSources.LookupTable import load_freesurfer_lut_file
import os
import numpy
import dicom


import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import nipy




def import_nifti(filename): 
    nip = nipy.load_image(filename)
    return nipy_to_occiput(nip)
    
    
def import_mask(filename, lookup_table_filename=None): 
    # Load file 
    nip = nipy.load_image(filename)
    occ = nipy_to_occiput(nip)
    occ.set_mask_flag(1) 
    
    # Load the lookup table. If not specified, try to load from file with the same name as 
    # the mask image file. 
    if lookup_table_filename == None: 
        f = []
        f.append(os.path.splitext(filename)[0]+'.lut')
        f.append(os.path.splitext(os.path.splitext(filename)[0])[0]+'.lut') #This includes .nii.gz files 
        for lookup_table_filename in f: 
            try: 
                lut = load_freesurfer_lut_file(lookup_table_filename) 
            except: 
                lut = None 
                
    else: 
        lut = load_freesurfer_lut_file(lookup_table_filename) 
    if lut  is not None: 
        occ.set_lookup_table(lut) 
    return occ


def import_dicom(search_path, extension='IMA'): 
    import dcmstack
    from glob import glob
    search_string = search_path + '/*.' + extension
    src_paths = glob(search_string)
    stacks = dcmstack.parse_and_stack(src_paths)
    images = []
    for key in stacks.keys():
        stack = stacks[key] 
        image = nifti_to_occiput(stack.to_nifti())
        images.append(image) 
    return images 



def import_dicom_series(path, files_start_with=None, files_end_with=None, exclude_files_end_with=['.dat','.txt','.py','.pyc','.nii','.gz'] ):
        """Rudimentary file to load dicom serie from a directory. """ 
        N=0 
        paths  = []
        slices = []
        files = os.listdir(path)
 
        for file_name in files: 
            file_valid = True
            if files_start_with is not None: 
                if not file_name.startswith(files_start_with): 
                    file_valid = False
            if files_end_with is not None: 
                if not file_name.endswith(files_end_with): 
                    file_valid = False 
            for s in exclude_files_end_with: 
                if file_name.endswith(s): 
                    file_valid = False           
            if file_valid: 
                full_path = path+os.sep+file_name
                # read moco information from files 
                paths.append(full_path)
                f = dicom.read_file(full_path) 
                slice = f.pixel_array
                slices.append(slice)
                N+=1
                instance_number    = f.get(0x00200013).value 
                creation_time      = f.get(0x00080013).value
                #print "Instance number:    ",instance_number
                #print "Creation time:      ",creation_time 
        array = numpy.zeros((slices[0].shape[0],slices[0].shape[1],N),dtype=numpy.float32)
        for i in range(N):  
            slice = numpy.float32(slices[i])  #FIXME: handle other data types 
            array[:,:,i] = slice 
        #return occiput_from_array(array)
        return array
    
